import os
import math
import csv

import gmsh
import json
import numpy as np
import pandas as pd
from pathlib import Path

from fiqus.geom_generators.GeometryCCT import Winding, FQPL
from fiqus.data.DataWindingsCCT import WindingsInformation
from fiqus.parsers.ParserDAT import ParserDAT
from fiqus.utils.Utils import GmshUtils, FilesAndFolders


class Post_Process:
    def __init__(self, fdm, verbose=True):
        """
        Class to cct models postprocessing
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.cctdm = fdm.magnet
        self.model_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name

        self.verbose = verbose
        self.gu = GmshUtils(self.model_folder, self.verbose)
        self.gu.initialize()
        self.masks_fqpls = {}
        self.pos_names = []
        for variable, volume, file_ext in zip(self.cctdm.postproc.variables, self.cctdm.postproc.volumes, self.cctdm.postproc.file_exts):
            self.pos_names.append(f'{variable}_{volume}.{file_ext}')
        self.pos_name = self.pos_names[0]
        self.field_map_3D = os.path.join(self.model_folder, 'field_map_3D.csv')
        self.model_file = self.field_map_3D
        self.geom_folder = Path(self.model_folder).parent.parent
        # csv output definition for fields
        self.csv_ds = {'ds [m]': None}  # length of each hexahedron
        self.csv_sAve = {'sAvePositions [m]': None}  # cumulative positions along the electrical order of the center of each hexahedron
        self.csv_3Dsurf = {  # basically this is the mapping of coordinates of corners to output. This dictionary defines which corner of hexahedra gets
                             # used and which coordinate for each name like 'x3Dsurf 1'. Look into Hexahedron.py base class for drawing of corner numbering
            'x3Dsurf 1 [m]': {5: 'x'},
            'x3Dsurf 2 [m]': {6: 'x'},
            'y3Dsurf 1 [m]': {5: 'y'},
            'y3Dsurf 2 [m]': {6: 'y'},
            'z3Dsurf 1 [m]': {5: 'z'},
            'z3Dsurf 2 [m]': {6: 'z'},
            'x3Dsurf_2 1 [m]': {7: 'x'},
            'x3Dsurf_2 2 [m]': {6: 'x'},
            'y3Dsurf_2 1 [m]': {7: 'y'},
            'y3Dsurf_2 2 [m]': {6: 'y'},
            'z3Dsurf_2 1 [m]': {7: 'z'},
            'z3Dsurf_2 2 [m]': {6: 'z'}}
        self.csv_nodeToTurns = {'nodeToHalfTurn [-]': None,  # hexahedron (node) mapping to "Electrical" turns (electrical order)
                                'nodeToPhysicalTurn [-]': None}  # hexahedron (node) mapping to "Physical" turns (this is used for defining thermal connections between turns)
        self.csv_TransportCurrent = {'TransportCurrent [A]': None}  # Transport current in the turn (not channel) for which the magnetic field was calculated.
        # The above does not need to be changed for running models at different current.
        self.csv_Bs = {
            # lists with magnetic field flux density components along cartesian coordinates (x, y, z) and l, h, w i.e. length, height and width of the channel (or wires if aligned with channel).
            # x, y, z is used for display, l, h, w are used in LEDET for Ic scaling (well only h and w for perpendicular field).
            'Bx [T]': 'Vx',
            'By [T]': 'Vy',
            'Bz [T]': 'Vz',
            'Bl [T]': 'Vl',
            'Bh [T]': 'Vh',
            'Bw [T]': 'Vw',
        }
        self.headers_dict = {**self.csv_ds, **self.csv_sAve, **self.csv_3Dsurf, **self.csv_nodeToTurns, **self.csv_TransportCurrent, **self.csv_Bs}
        self.data_for_csv = {}
        self.precision_for_csv = {}
        # -------- make keys in data for csv ----
        for key in list(self.headers_dict.keys()):
            self.precision_for_csv[key] = '%.6f'  # data will be written to csv output file with 6 decimal places, i.e. down to um and uT level.
        for key in list(self.csv_ds.keys()):
            self.data_for_csv[key] = []
        for key in list(self.csv_sAve.keys()):
            self.data_for_csv[key] = []
        for key in list(self.csv_3Dsurf.keys()):
            self.data_for_csv[key] = []
        for key in list(self.csv_nodeToTurns.keys()):
            self.data_for_csv[key] = []
        for key in list(self.csv_TransportCurrent.keys()):
            self.data_for_csv[key] = []
        for key in list(self.csv_Bs.keys()):
            self.data_for_csv[key] = []
        self.distance_key = list(self.csv_ds.keys())[0]
        self.sAve_key = list(self.csv_sAve.keys())[0]

    @staticmethod
    def _get_fields(coord_center, normals):
        """
        Helper funciton to probe magnetic field solution form gmsh view and calculate magnetic field along the height, width and length of the channel section (hex element)
        :param coord_center: dictionary with 'x', 'y' and 'z' coordinates stored as lists with items for each hexahedra. Magnetic field is probed at these locations.
        :param normals:  dictionary with normals along the height, width and length stored per coordinate direction 'n_x', 'n_y' and 'n_z' and lists with items for each hexahedra.
        :return: tuple with list of Bxs, Bys, Bzs, Bhs, Bws, Bls
        """
        Bxs = []
        Bys = []
        Bzs = []
        Bhs = []
        Bws = []
        Bls = []
        view_tag = gmsh.view.getTags()[0]
        for v, _ in enumerate(coord_center['x']):
            field = gmsh.view.probe(view_tag, coord_center['x'][v], coord_center['y'][v], coord_center['z'][v])[0]
            Bx = field[0]
            By = field[1]
            Bz = field[2]
            Bmod = math.sqrt(Bx ** 2 + By ** 2 + Bz ** 2)
            Bh = abs(Bx * normals['normals_h']['n_x'][v] + By * normals['normals_h']['n_y'][v] + Bz * normals['normals_h']['n_z'][v])
            Bw = abs(Bx * normals['normals_w']['n_x'][v] + By * normals['normals_w']['n_y'][v] + Bz * normals['normals_w']['n_z'][v])
            Bl = math.sqrt(abs(Bmod ** 2 - Bh ** 2 - Bw ** 2))  # sometimes it is very small and negative, so force it to be positive before taking the root.
            for arr, val in zip([Bxs, Bys, Bzs, Bhs, Bws, Bls], [Bx, By, Bz, Bh, Bw, Bl]):
                arr.append(val)
        return Bxs, Bys, Bzs, Bhs, Bws, Bls

    @staticmethod
    def _plot_fields_in_views(p_name, global_channel_pos, coord_center, Bxs, Bys, Bzs):
        """
        Add gmsh list data view with name 'B cartesian {p_name}, {global_channel_pos}' at coordinates stored in coord_center and magnetic field values from list Bxs, Bys, Bzs
        :param p_name: string with powered region name
        :param global_channel_pos: integer with wire position in the channel
        :param coord_center: dictionary with 'x', 'y' and 'z' coordinates stored as lists with items for each hexahedra. Magnetic field is plotted at these locations.
        :param Bxs: list of magnetic field along the x axis to use in the view
        :param Bys: list of magnetic field along the y axis to use in the view
        :param Bzs: list of magnetic field along the z axis to use in the view
        :return: none, adds view to currently initialized gmsh model and synchronizes
        """
        data_cartesian = []
        for v, _ in enumerate(coord_center['x']):
            data_cartesian.append(coord_center['x'][v])
            data_cartesian.append(coord_center['y'][v])
            data_cartesian.append(coord_center['z'][v])
            data_cartesian.append(Bxs[v])
            data_cartesian.append(Bys[v])
            data_cartesian.append(Bzs[v])
        gmsh.view.addListData(gmsh.view.add(f'B cartesian {p_name}, {global_channel_pos}'), "VP", len(data_cartesian) // 6, data_cartesian)
        gmsh.model.occ.synchronize()

    @staticmethod
    def _plot_turns_in_view(data_turns, coord_center, turns_for_ch_pos):
        """
        Add gmsh list data view with name 'Turn numbering' at coordinates stored in coord_center and values from turns_for_ch_pos
        :param data_turns: This is list to which the data gets appended so at the end one view with all the turns labeled is created with looping through turns in the channels.
        :param coord_center: dictionary with 'x', 'y' and 'z' coordinates stored as lists with items for each hexahedra. Magnetic field is plotted at these locations.
        :param turns_for_ch_pos: Lists with turn numbers to plot in the view.
        :return: none, adds view to currently initialized gmsh model and synchronizes
        """
        for v, _ in enumerate(coord_center['x']):
            data_turns.append(coord_center['x'][v])
            data_turns.append(coord_center['y'][v])
            data_turns.append(coord_center['z'][v])
            data_turns.append(turns_for_ch_pos[v])
        gmsh.view.addListData(gmsh.view.add('Turn numbering'), "SP", len(data_turns) // 4, data_turns)
        gmsh.model.occ.synchronize()

    @staticmethod
    def _check_normals_match_corners(coord_center, normals_dict, p_name):
        """
        This is error checking function. Coord_centers are calculated in this class 'on the 'fly' form cctdm file. Normals are generated form geom_generators before meshing, then saved to file and now loaded here.
        There is a chance that thses could get 'out of sync'.         A basic check of the length (i.e. number of hexahedra involved is performed) of these lists is performed.
        This will only detect if number of turns or discretization of each powered geom_generators has changed.
        :param coord_center: list with coordinate centers. The content is not used only list length is queried.
        :param normals_dict: list with normals read form json file. The content is not used only list length is queried.
        :param p_name: string with powered volume name. Only used for error message display to say which region is not matching.
        :return: None, prints error to the python console, if any.
        """
        for cord in ['x', 'y', 'z']:
            if coord_center[cord].size != len(normals_dict['normals_h'][cord]) or coord_center[cord].size != len(normals_dict['normals_w'][cord]):
                raise ValueError(f'Number of volumes in normals does not match the magnet definition for {p_name} winding!')

    @staticmethod
    def _calc_coord_center_and_ds(corners_lists):
        """
        Calculates a geometrical centre of each hexahedron base on its corners coordinates
        :param corners_lists: lists of lists corresponding to a list of hexahedrons with each list containing 8 corner coordinates for each of the 8 points in the corners of hexahedron
        :return: coord_center: dictionary with 'x', 'y' and 'z' coordinates stored as lists with items for each hexahedra. and ds
        """
        coord_center = {}
        for cord in ['x', 'y', 'z']:
            coord_center[cord] = np.mean(corners_lists[cord], axis=0)  # coordinates center
        surf_close = {}
        surf_far = {}
        for cord in ['x', 'y', 'z']:
            surf_close[cord] = np.mean(corners_lists[cord][0:4], axis=0)  # coordinates surface close
            surf_far[cord] = np.mean(corners_lists[cord][4:8], axis=0)  # coordinates surface far
        ds = np.sqrt(np.square(surf_close['x'] - surf_far['x']) + np.square(surf_close['y'] - surf_far['y']) + np.square(surf_close['z'] - surf_far['z']))
        return coord_center, ds

    def _load_to_fields_csv_ds(self, ds, h3Dsurf, nodeToHalfTurn, nodeToPhysicalTurn, current, Bxs, Bys, Bzs, Bls, Bhs, Bws, s_ave):
        """
        Method to load data to csv dictionary to be dumped to a csv file.
        :param ds: length of each hexahedron
        :param h3Dsurf: dict with coordinates to export
        :param nodeToHalfTurn: hexahedron (node) mapping to "Electrical" turns (electrical order)
        :param nodeToPhysicalTurn: hexahedron (node) mapping to "Physical" turns (this is used for defining thermal connections between turns)
        :param current: Transport current in the turn (not channel) for which the magnetic field was calculated. It does not need to be changed for running models at different current.
        :param Bxs: list of magnetic flux density component along x direction
        :param Bys: list of magnetic flux density component along y direction
        :param Bzs: list of magnetic flux density component along z direction
        :param Bls: list of magnetic flux density component along l (length) of wire direction
        :param Bhs: list of magnetic flux density component along h (height) of wire direction
        :param Bws: list of magnetic flux density component along w (wight) of wire direction
        :param s_ave: cumulative positions along the electrical order of the center of each hexahedron
        :return: nothing, appends data to this class attribute data_for_csv
        """
        for arr, key in zip([ds, s_ave], list(self.csv_ds.keys()) + list(self.csv_sAve.keys())):
            self.data_for_csv[key].extend(arr)
        for key, dict_what in self.csv_3Dsurf.items():
            for p_n, p_c in dict_what.items():
                self.data_for_csv[key].extend(h3Dsurf[p_c][p_n])
        for arr, key in zip([Bxs, Bys, Bzs, Bls, Bhs, Bws, current, nodeToHalfTurn, nodeToPhysicalTurn], list(self.csv_Bs.keys()) + list(self.csv_TransportCurrent.keys()) + list(self.csv_nodeToTurns.keys())):
            self.data_for_csv[key].extend(arr)

    def _save_fields_csv(self):
        """
        Helper method to save csv file from data_for_csv class attribute.
        :return: nothing, saves csv to file with name magnet_name(from yaml input file)_total_n_turns.csv
        """
        # csv_file_path = os.path.join(self.model_folder, f'{self.magnet_name[:self.magnet_name.index("_")]}_{int(total_n_turns)}.csv')

        df = pd.DataFrame(self.data_for_csv)
        for column in df:
            df[column] = df[column].map(lambda x: self.precision_for_csv[column] % x)
        # put NaN at the end of some columns to match what LEDET needs.
        df.loc[df.index[-1], self.distance_key] = 'NaN'
        df.loc[df.index[-1], self.sAve_key] = 'NaN'
        # df.loc[df.index[-1], list(self.csv_nodeToTurns.keys())[0]] = 'NaN'
        for h in list(self.csv_nodeToTurns.keys()):
            df.loc[df.index[-1], h] = 'NaN'
        df.loc[df.index[-1], list(self.csv_TransportCurrent.keys())[0]] = 'NaN'
        for h in list(self.csv_Bs.keys()):
            df.loc[df.index[-1], h] = 'NaN'
        df.to_csv(self.field_map_3D, index=False, sep=',')  # , float_format='%.7f')

    @staticmethod
    def _trim_array(array, mask):
        return list(np.array(array)[mask])

    @staticmethod
    def _all_equal(iterator):
        iterator = iter(iterator)
        try:
            first = next(iterator)
        except StopIteration:
            return True
        return all(first == x for x in iterator)

    def postporcess_inductance(self):
        """
        Function to postprocess .dat file with inductance saved by GetDP to selfMutualInductanceMatrix.csv required by LEDET for CCT magnet solve
        :return: Nothing, writes selfMutualInductanceMatrix.csv file on disc.
        :rtype: None
        """
        inductance_file_from_getdp = os.path.join(self.model_folder, 'Inductance.dat')
        channels_inductance = ParserDAT(inductance_file_from_getdp).pqv    # postprocessed quantity value (pqv) for Inductance.dat contains magnet self-inductance  (channel, not wire turns)

        total_n_turns_channel = 0   # including fqpls
        for n_turns in self.cctdm.geometry.windings.n_turnss:
            total_n_turns_channel = total_n_turns_channel + n_turns
        if self._all_equal(self.cctdm.postproc.windings_wwns) and self._all_equal(self.cctdm.postproc.windings_whns):
            number_of_turns_in_channel = self.cctdm.postproc.windings_wwns[0] * self.cctdm.postproc.windings_whns[0]
        total_n_turns_channel = int(total_n_turns_channel)
        total_n_turns_channel_and_fqpls = total_n_turns_channel

        #windings_inductance = channels_inductance * total_n_turns_windings**2
        windings_inductance = channels_inductance * number_of_turns_in_channel**3   # scaling with square for number of turns in the channel as the model has current in the channel but not number of turns


        geometry_folder = Path(self.model_folder).parent.parent
        winding_information_file = os.path.join(geometry_folder, f"{self.magnet_name}.wi")
        cctwi = FilesAndFolders.read_data_from_yaml(winding_information_file, WindingsInformation)
        windings_avg_length = cctwi.windings_avg_length
        windings_inductance_per_m = windings_inductance / windings_avg_length

        print(f"Channels self-inductance: {channels_inductance} H")
        print(f"Number of turns in channel: {number_of_turns_in_channel} H")
        print(f"Total number of channel turns: {total_n_turns_channel} turns")
        print(f"Total number of inductance blocks: {total_n_turns_channel}")
        print(f"Windings self-inductance: {windings_inductance} H")
        print(f"Windings self-inductance per meter: {windings_inductance_per_m} H/m")
        print(f"Magnetic length: {windings_avg_length} m")

        fqpls_dummy_inductance_block = 1e-9    # H/m   (H per meter of magnet!)
        for _ in self.cctdm.geometry.fqpls.names:
            total_n_turns_channel_and_fqpls = total_n_turns_channel_and_fqpls + 1
            windings_inductance_per_m = windings_inductance_per_m - fqpls_dummy_inductance_block        # subtracting fqpls as they will be added later to the matrix

        win_ind_for_matrix = windings_inductance_per_m / total_n_turns_channel**2
        M_matrix = np.ones(shape=(total_n_turns_channel_and_fqpls, total_n_turns_channel_and_fqpls)) * win_ind_for_matrix
        M_matrix[:, total_n_turns_channel:total_n_turns_channel_and_fqpls] = 0.0       # mutual inductance of windings to fqpl set to zero for the columns
        M_matrix[total_n_turns_channel:total_n_turns_channel_and_fqpls, :] = 0.0       # mutual inductance of windings to fqpl set to zero for the rows
        for ij in range(total_n_turns_channel, total_n_turns_channel_and_fqpls, 1):
            M_matrix[ij, ij] = fqpls_dummy_inductance_block                                         # self-inductance of fqpls inductance block
        # below code is the same as in BuilderLEDET in steam_sdk
        csv_write_path = os.path.join(self.model_folder, 'selfMutualInductanceMatrix.csv')
        print(f'Saving square matrix of size {total_n_turns_channel_and_fqpls}x{total_n_turns_channel_and_fqpls} into: {csv_write_path}')
        with open(csv_write_path, 'w', newline='') as file:
            reader = csv.writer(file)
            reader.writerow(["# Extended self mutual inductance matrix [H/m]"])
            for i in range(M_matrix.shape[0]):
                reader.writerow(M_matrix[i])

    def postprocess_fields(self, gui=False):
        """
        Methods to calculated 'virtual' turn positions in the channels and output information for about them for LEDET as a csv file.
        :param gui: if True, graphical user interface (gui) of gmsh is shown with views created
        :return: nothing, writes csv to file
        """
        winding_order = self.cctdm.postproc.winding_order
        total_n_turns = 0
        for n_turns, wwn, whn in zip(self.cctdm.geometry.windings.n_turnss, self.cctdm.postproc.windings_wwns, self.cctdm.postproc.windings_whns):
            total_n_turns = total_n_turns + n_turns * wwn * whn

        gmsh.open(self.pos_name)
        global_channel_pos = 1
        # data_turns = []  # used in self._plot_turns_in_view() - do not delete
        csv_data_dict = {}
        s_max = 0
        for w_i, w_name in enumerate(self.cctdm.geometry.windings.names):  # + self.cctwi.f_names:
            normals_dict = json.load(open(os.path.join(self.geom_folder, f'{w_name}.normals')))
            winding_obj = Winding(self.cctdm, w_i, post_proc=True)
            ww2 = self.cctdm.geometry.windings.wwws[w_i] / 2  # half of channel width
            wh2 = self.cctdm.geometry.windings.wwhs[w_i] / 2  # half of channel height
            wwns = self.cctdm.postproc.windings_wwns[w_i]  # number of wires in width direction
            whns = self.cctdm.postproc.windings_whns[w_i]  # number of wires in height direction
            wsw = self.cctdm.geometry.windings.wwws[w_i] / wwns  # wire size in width
            wsh = self.cctdm.geometry.windings.wwhs[w_i] / whns  # wire size in height
            for i_w in range(wwns):
                for i_h in range(whns):
                    corner_i = 0
                    corners_lists = {'x': [], 'y': [], 'z': []}
                    for far in [False, True]:
                        for ii_w, ii_h in zip([0, 0, 1, 1], [0, 1, 1, 0]):
                            wt = (-ww2 + (ii_w + i_w) * wsw) / ww2
                            ht = (-wh2 + (ii_h + i_h) * wsh) / wh2
                            corner_i += 1
                            xs, ys, zs, turns_from_rotation = winding_obj.calc_turns_corner_coords(wt, ht, far=far)
                            zs = [z - winding_obj.z_corr for z in zs]
                            corners_lists['x'].append(xs)
                            corners_lists['y'].append(ys)
                            corners_lists['z'].append(zs)
                    nodeToPhysicalTurn = [(global_channel_pos - 1) * self.cctdm.geometry.windings.n_turnss[w_i] + t for t in turns_from_rotation]
                    nodeToHalfTurn = [winding_order.index(global_channel_pos) * self.cctdm.geometry.windings.n_turnss[w_i] + t for t in turns_from_rotation]
                    coord_center, ds = self._calc_coord_center_and_ds(corners_lists)
                    self._check_normals_match_corners(coord_center, normals_dict, w_name)
                    # if global_channel_pos == 1:
                    #     s_ave_elem_len = np.copy(ds)
                    #     s_ave_elem_len[0] = s_ave_elem_len[0] / 2
                    #     s_ave = np.cumsum(s_ave_elem_len)
                    # else:
                    #     s_ave = np.cumsum(ds) + s_max
                    # s_max = np.max(s_ave)
                    current = [abs(self.cctdm.solve.windings.currents[w_i]) / (wwns * whns)] * len(nodeToHalfTurn)
                    Bxs, Bys, Bzs, Bhs, Bws, Bls = self._get_fields(coord_center, normals_dict)
                    if gui:
                        self._plot_fields_in_views(w_name, global_channel_pos, coord_center, Bxs, Bys, Bzs)
                        # self._plot_turns_in_view(data_turns, coord_center, nodeToHalfTurn)
                    csv_data_dict[winding_order.index(global_channel_pos) + 1] = ds, corners_lists, nodeToHalfTurn, nodeToPhysicalTurn, current, Bxs, Bys, Bzs, Bls, Bhs, Bws
                    global_channel_pos += 1
        for f_i, f_name in enumerate(self.cctdm.geometry.fqpls.names):
            # ----- calculate centers of coordinates for hexes of fqpl
            fqpl_obj = FQPL(self.cctdm, f_i)
            corners_lists = {'x': [], 'y': [], 'z': []}
            for corner_i in range(8):
                for cord in ['x', 'y', 'z']:
                    corners_lists[cord].append([v_dict[corner_i][cord] for v_dict in fqpl_obj.hexes.values()])
            coord_center, ds = self._calc_coord_center_and_ds(corners_lists)
            normals_dict = json.load(open(os.path.join(self.geom_folder, f'{f_name}.normals')))
            self._check_normals_match_corners(coord_center, normals_dict, f_name)
            Bxs, Bys, Bzs, Bhs, Bws, Bls = self._get_fields(coord_center, normals_dict)
            self.masks_fqpls[f_i] = []          # dictionary to keep masks for trimming one end of fqpls that is long to extend to the end of air region, but not needed that long in simulations
            for z_close, z_far in zip(corners_lists['z'][0], corners_lists['z'][-1]):
                if z_close > -self.cctdm.geometry.fqpls.z_ends[f_i] * self.cctdm.postproc.fqpl_export_trim_tol[f_i] and z_far > -self.cctdm.geometry.fqpls.z_ends[f_i] * self.cctdm.postproc.fqpl_export_trim_tol[f_i]:
                    self.masks_fqpls[f_i].append(True)
                else:
                    self.masks_fqpls[f_i].append(False)
            self.masks_fqpls[f_i] = np.array(self.masks_fqpls[f_i], dtype=bool)
            for cord in ['x', 'y', 'z']:
                for corner_i, corner in enumerate(corners_lists[cord]):
                    corners_lists[cord][corner_i] = list(np.array(corner)[self.masks_fqpls[f_i]])
            lists_to_trim = [Bxs, Bys, Bzs, Bhs, Bws, Bls, ds]
            for idx, array in enumerate(lists_to_trim):
                lists_to_trim[idx] = self._trim_array(array, self.masks_fqpls[f_i])
            [Bxs, Bys, Bzs, Bhs, Bws, Bls, ds] = lists_to_trim
            for cord in ['x', 'y', 'z']:
                coord_center[cord] = self._trim_array(coord_center[cord], self.masks_fqpls[f_i])
            # s_ave = np.cumsum(ds) + s_max
            # s_max = np.max(s_ave)
            go_dict = {'from': 0, 'to': len(Bxs)//2}       # fqpl 'go' part
            ret_dict = {'from': len(Bxs)//2, 'to': len(Bxs)}
            for go_ret in [go_dict, ret_dict]:
                total_n_turns += 1
                ds_half = ds[go_ret['from']:go_ret['to']]
                corners_lists_half = {}
                for cord in ['x', 'y', 'z']:
                    corners_lists_half[cord] = []
                    for corner_i, corner in enumerate(corners_lists[cord]):
                        corners_lists_half[cord].append(corner[go_ret['from']:go_ret['to']])
                nodeToHalfTurn = [total_n_turns] * len(Bxs[go_ret['from']:go_ret['to']])  # total expected number of turns + one + fqpl number
                nodeToPhysicalTurn = [total_n_turns] * len(Bxs[go_ret['from']:go_ret['to']])
                current = [self.cctdm.solve.fqpls.currents[f_i]] * len(Bxs[go_ret['from']:go_ret['to']])
                Bxs_half = Bxs[go_ret['from']:go_ret['to']]
                Bys_half = Bys[go_ret['from']:go_ret['to']]
                Bzs_half = Bzs[go_ret['from']:go_ret['to']]
                Bls_half = Bls[go_ret['from']:go_ret['to']]
                Bhs_half = Bhs[go_ret['from']:go_ret['to']]
                Bws_half = Bws[go_ret['from']:go_ret['to']]
                #s_ave_half = s_ave[go_ret['from']:go_ret['to']]
                csv_data_dict[total_n_turns] = ds_half, corners_lists_half, nodeToHalfTurn, nodeToPhysicalTurn, current, Bxs_half, Bys_half, Bzs_half, Bls_half, Bhs_half, Bws_half
            if gui:
                self._plot_fields_in_views(f_name, global_channel_pos, coord_center, Bxs, Bys, Bzs)
            global_channel_pos += 1

        ds_global = []
        csv_data_sorded = dict(sorted(csv_data_dict.items())).values()# sorted list of tuples with the csv data
        for tuple_with_values in csv_data_sorded:
            ds_global.append(tuple_with_values[0])      # 0 index in tuple is for ds, changing to list to easily extend.

        s_ave_global = []
        s_max = -ds_global[0][0]/2 # make a s_max to a negative half of the length of the first hexahedra. This is to set the first s_ave max to a half of the first ds element
        for ds_i, ds_np in enumerate(ds_global):
            s_ave_np = np.cumsum(ds_np) + s_max
            s_ave_global.append(s_ave_np)
            s_max = np.max(s_ave_np)

        for sorted_v_tuple, s_ave_np in zip(csv_data_sorded, s_ave_global):
            self._load_to_fields_csv_ds(*sorted_v_tuple, s_ave_np)  # this is to avoid rearranging s_ave by electrical order

        #el_order_half_turns = [int(t) for t in list(pd.unique(self.data_for_csv['nodeToPhysicalTurn [-]']))]
        #json.dump({'el_order_half_turns': el_order_half_turns}, open(os.path.join(self.model_folder, "el_order_half_turns.json"), 'w'))
        self._save_fields_csv()
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()




from pydantic import BaseModel
from typing import List


class Region(BaseModel):
    name: str = None
    number: int = None


class Regions(BaseModel):
    names: List[str] = None
    numbers: List[int] = None


class PoweredRegions(BaseModel):
    names: List[str] = None
    numbers: List[int] = None
    currents: List[float] = None
    sigmas: List[float] = None
    mu_rs: List[float] = None


class InducedRegions(BaseModel):
    names: List[str] = None
    numbers: List[int] = None
    sigmas: List[float] = None
    mu_rs: List[float] = None


class IronRegions(BaseModel):
    names: List[str] = None
    numbers: List[int] = None
    sigmas: List[float] = None
    mu_rs: List[float] = None


class AirRegion(BaseModel):
    name: str = None
    number: int = None
    sigma: float = None
    mu_r: float = None


class AirFarFieldRegion(BaseModel):
    name: str = None
    number: int = None
    radius_in: float = None
    radius_out: float = None


class Powered(BaseModel):
    vol: PoweredRegions = PoweredRegions()  # volume region
    surf_in: Regions = Regions()  # input terminal surface region
    surf_out: Regions = Regions()  # output terminal surface region
    cochain: Regions = Regions()  # winding cochain (cut)


class Induced(BaseModel):
    vol: InducedRegions = InducedRegions()  # volume region
    cochain: Regions = Regions()  # winding cochain (cut)


class Iron(BaseModel):
    vol: IronRegions = IronRegions()  # volume region
    surf: Regions = Regions()  # surface region


class Air(BaseModel):
    vol: AirRegion = AirRegion()  # volume region
    surf: Region = Region()  # surface region
    line: Region = Region()    # line region


class AirFarField(BaseModel):
    vol: AirFarFieldRegion = AirFarFieldRegion()  # volume region
    surf: Region = Region()  # surface region


class RegionsModel(BaseModel):
    powered: Powered = Powered()
    induced: Induced = Induced()
    iron: Iron = Iron()
    air: Air = Air()
    air_far_field: AirFarField = AirFarField()


# if __name__ == "__main__":
#     write = True
#     read = False
#
#     def read_regions(regions_file_name):
#         with open(regions_file_name, 'r') as stream:
#             yaml_str = ruamel.yaml.safe_load(stream)
#         return RegionsModel(**yaml_str)
#
#     def flist(x):
#         retval = ruamel.yaml.comments.CommentedSeq(x)
#         retval.fa.set_flow_style()  # fa -> format attribute
#         return retval
#
#     if write:
#         model = RegionsModel()
#         model.powered.vol = [1, 2]
#         data_dict = model.dict()
#         yaml = ruamel.yaml.YAML()
#         yaml.default_flow_style = False
#         yaml.emitter.alt_null = 'Null'
#
#         def my_represent_none(self, data):
#             return self.represent_scalar('tag:yaml.org,2002:null', 'null')
#
#         yaml.representer.add_representer(type(None), my_represent_none)
#         with open('cct_regions_empty.yaml', 'w') as yaml_file:
#             yaml.dump(model.dict(), yaml_file)
#     if read:
#         regions_file_name = 'cct1_regions_manual.yaml'
#         regions = read_regions(regions_file_name)



import os
import gmsh
from numpy import square as square
from numpy import sqrt as sqrt

from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.data import DataMultipole as dM
from fiqus.data.RegionsModelFiQuS import RegionsModel as rM


class Mesh:
    def __init__(self, data: dF.FDM() = None, sett: dF.FiQuSSettings() = None, mesh_folder: str = None,
                 verbose: bool = False):
        """
        Class to generate mesh
        :param data: FiQuS data model
        :param sett: settings data model
        :param verbose: If True more information is printed in python console.
        """
        self.data: dF.FDM() = data
        self.set: dF.FiQuSSettings() = sett
        self.mesh_folder = mesh_folder
        self.verbose: bool = verbose

        self.md = dM.MultipoleData()
        self.rm = rM()

        self.gu = GmshUtils(self.mesh_folder, self.verbose)
        self.gu.initialize()
        self.occ = gmsh.model.occ
        self.mesh = gmsh.model.mesh

        self.brep_iron_curves = {1: set(), 2: set(), 3: set(), 4: set()}
        self.mesh_parameters = dict.fromkeys(['SJ', 'SICN', 'SIGE', 'Gamma', 'nodes'])
        self.geom_folder = os.path.dirname(self.mesh_folder)
        self.model_file = f"{os.path.join(self.mesh_folder, self.data.general.magnet_name)}.msh"

        self.colors = {'wedges': [86, 180, 233],  # sky blue
                       'half_turns_pos': [213, 94, 0],  # vermilion
                       'half_turns_neg': [255, 136, 42],  # light vermilion
                       'air': [240, 228, 66],  # yellow
                       'air_inf': [220, 208, 46],  # dark yellow
                       # yoke
                       'BHiron1': [0, 114, 178],  # blue
                       'BHiron2': [0, 158, 115],  # bluish green
                       'BHiron4': [86, 180, 233],  # sky blue
                       # key
                       'BHiron3': [220, 208, 46],  # dark yellow
                       # [230, 159, 0],  # orange
                       # collar
                       'BHiron5': [204, 121, 167],  # hopbush
                       'BHiron6': [0, 114, 178],  # blue
                       'BHiron7': [204, 121, 167]}  # reddish purple

    def ending_step(self, gui: bool = False):
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()

    def loadAuxiliaryFile(self):
        self.md = Util.read_data_from_yaml(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.aux", dM.MultipoleData)

    def updateAuxiliaryFile(self):
        md2 = Util.read_data_from_yaml(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.aux", dM.MultipoleData)
        md2.domains.physical_groups = self.md.domains.physical_groups
        Util.write_data_to_yaml(f"{os.path.join(self.geom_folder, self.data.general.magnet_name)}.aux", md2.dict())

    def saveMeshFile(self):
        gmsh.write(self.model_file)

    def saveRegionFile(self):
        Util.write_data_to_yaml(f"{os.path.join(self.mesh_folder, self.data.general.magnet_name)}.reg", self.rm.dict())

    def getIronCurvesTags(self):
        if self.data.magnet.geometry.with_iron_yoke:
            for quadrant, qq in self.md.geometries.iron.quadrants.items():
                for area_name, area in qq.areas.items():
                    if area.surface:
                        self.brep_iron_curves[quadrant] |= set(gmsh.model.getAdjacencies(2, area.surface)[1])

    def defineMesh(self):
        curves_list = []
        for coil_nr, coil in self.md.geometries.coil.coils.items():
            for pole_nr, pole in coil.poles.items():
                for layer_nr, layer in pole.layers.items():
                    for winding_nr, winding in layer.windings.items():
                        for block_key, block in winding.blocks.items():
                            hts = block.half_turns
                            for i in range(len(hts.areas) - 1):
                                curves_list.append(self.occ.copy([(1, hts.lines[str(i + 1) + 'r'])])[0][1])
                            curves_list.append(self.occ.copy([(1, hts.lines['end'])])[0][1])

        self.occ.synchronize()
        if not self.data.magnet.mesh.default_mesh:
            curve_list_iron = []
            if self.data.magnet.geometry.with_iron_yoke:
                # if self.from_brep:
                for quadrant, qq in self.brep_iron_curves.items():
                    for line in qq:
                        curve_list_iron.append(self.occ.copy([(1, line)])[0][1])
                # else:
                #     for quadrant, qq in self.md.geometry.iron.quadrants.items():
                #         for line_name, line in qq.lines.items():
                #             curve_list_iron.append(self.occ.copy([(1, line)])[0][1])

            size_min = self.data.magnet.mesh.mesh_coil.SizeMin
            size_max = self.data.magnet.mesh.mesh_coil.SizeMax
            dist_min = self.data.magnet.mesh.mesh_coil.DistMin
            dist_max = self.data.magnet.mesh.mesh_coil.DistMax  # iron max_radius

            list_aux_type = "CurvesList"
            list_aux = curve_list_iron
            size_min_aux = self.data.magnet.mesh.mesh_iron.SizeMin
            size_max_aux = self.data.magnet.mesh.mesh_iron.SizeMax
            dist_min_aux = self.data.magnet.mesh.mesh_iron.DistMin
            dist_max_aux = self.data.magnet.mesh.mesh_iron.DistMax

        else:
            min_height = 1.
            for name, cond in self.set.Model_Data_GS.conductors.items():
                min_height = min(min_height, cond.cable.bare_cable_height_mean)

            bore_centers = [self.occ.addPoint(coil.bore_center.x, coil.bore_center.y, 0.)
                            for coil_nr, coil in self.md.geometries.coil.coils.items()]
            point = gmsh.model.getValue(0, self.md.geometries.coil.coils[1].poles[1].layers[1].windings[1].blocks[
                1].half_turns.points['1i'], [])
            min_dist = sqrt(square(point[0] - self.md.geometries.coil.coils[1].bore_center.x) +
                            square(point[1] - self.md.geometries.coil.coils[1].bore_center.y))

            size_min = min_height / 2
            size_max = min_height * 20
            dist_min = min_height * 2
            dist_max = self.md.geometries.iron.max_radius

            list_aux_type = "PointsList"
            list_aux = bore_centers
            size_min_aux = min_height / 2
            size_max_aux = min_height * 20
            dist_min_aux = min_dist
            dist_max_aux = min_dist

        self.occ.synchronize()

        distance_coil = self.mesh.field.add("Distance")
        self.mesh.field.setNumbers(distance_coil, "CurvesList", curves_list)
        self.mesh.field.setNumber(distance_coil, "Sampling", 100)

        threshold = self.mesh.field.add("Threshold")
        self.mesh.field.setNumber(threshold, "InField", distance_coil)
        self.mesh.field.setNumber(threshold, "SizeMin", size_min)
        self.mesh.field.setNumber(threshold, "SizeMax", size_max)
        self.mesh.field.setNumber(threshold, "DistMin", dist_min)
        self.mesh.field.setNumber(threshold, "DistMax", dist_max)

        distance_aux = self.mesh.field.add("Distance")
        self.mesh.field.setNumbers(distance_aux, list_aux_type, list_aux)
        self.mesh.field.setNumber(distance_aux, "Sampling", 100)

        threshold_aux = self.mesh.field.add("Threshold")
        self.mesh.field.setNumber(threshold_aux, "InField", distance_aux)
        self.mesh.field.setNumber(threshold_aux, "SizeMin", size_min_aux)
        self.mesh.field.setNumber(threshold_aux, "SizeMax", size_max_aux)
        self.mesh.field.setNumber(threshold_aux, "DistMin", dist_min_aux)
        self.mesh.field.setNumber(threshold_aux, "DistMax", dist_max_aux)

        background = self.mesh.field.add("Min")
        self.mesh.field.setNumbers(background, "FieldsList", [threshold, threshold_aux] if (
                self.data.magnet.geometry.with_iron_yoke and not self.data.magnet.mesh.default_mesh
                or self.data.magnet.mesh.default_mesh) else [threshold])
        self.mesh.field.setAsBackgroundMesh(background)

    def fragment(self):
        """
            Fragment and group air domains
        """
        holes = []
        for group_name, surfaces in self.md.domains.groups_surfaces.items():
            if group_name != 'air_inf':
                holes.extend([(2, s) for s in surfaces])
        fragmented = self.occ.fragment([(2, self.md.geometries.air_inf.areas['inner'].surface)], holes)[1]
        self.md.domains.groups_surfaces['air'] = []
        existing_domains = [e[0][1] for e in fragmented[1:]]
        for e in fragmented[0]:
            if e[1] not in existing_domains:
                self.md.domains.groups_surfaces['air'].append(e[1])

        self.occ.synchronize()

    def createPhysicalGroups(self):
        """
            Creates physical groups by grouping the mirrored entities according to the Roxie domains
        """
        pg = self.md.domains.physical_groups
        for group_name, surfaces in self.md.domains.groups_surfaces.items():
            pg.surfaces[group_name] = gmsh.model.addPhysicalGroup(2, surfaces)
            gmsh.model.setPhysicalName(2, pg.surfaces[group_name], group_name)
            if group_name[:5] == 'block':
                color = self.colors['half_turns_pos'] if group_name[-3:] == 'pos' else self.colors['half_turns_neg']
            elif group_name[:5] == 'wedge':
                color = self.colors['wedges']
            else:
                color = self.colors[group_name]
            gmsh.model.setColor([(2, i) for i in surfaces], color[0], color[1], color[2])

        pg.curves['air_inf'] = gmsh.model.addPhysicalGroup(1, [self.md.geometries.air_inf.lines['outer']])
        gmsh.model.setPhysicalName(1, pg.curves['air_inf'], 'air_inf')

    def assignRegionsTags(self):
        self.rm.air_far_field.vol.radius_out = float(gmsh.model.getValue(0, gmsh.model.getAdjacencies(
                1, self.md.geometries.air_inf.lines['outer'])[1][0], []).max())
        self.rm.air_far_field.vol.radius_in = float(gmsh.model.getValue(0, gmsh.model.getAdjacencies(
                1, self.md.geometries.air_inf.lines['inner'])[1][0], []).max())

        self.rm.air.vol.name = "Air"
        self.rm.air.vol.number = self.md.domains.physical_groups.surfaces['air']
        self.rm.air_far_field.vol.name = "AirInf"
        self.rm.air_far_field.vol.number = self.md.domains.physical_groups.surfaces['air_inf']
        self.rm.powered.vol.names = []
        self.rm.powered.vol.numbers = []
        self.rm.iron.vol.names = []
        self.rm.iron.vol.numbers = []
        self.rm.induced.vol.names = []
        self.rm.induced.vol.numbers = []
        for group_name, surface in self.md.domains.physical_groups.surfaces.items():
            if group_name[:5] == 'block':
                self.rm.powered.vol.names.append(group_name)
                self.rm.powered.vol.numbers.append(surface)
                #  self.fd.half_turns.cross_sections.append(self.conductor_areas[name])
            elif group_name[:2] == 'BH':
                self.rm.iron.vol.names.append(group_name)
                self.rm.iron.vol.numbers.append(surface)
            elif group_name[:5] == 'wedge':
                self.rm.induced.vol.names.append(group_name)
                self.rm.induced.vol.numbers.append(surface)

        self.rm.air_far_field.surf.name = "Surface_Inf"
        self.rm.air_far_field.surf.number = self.md.domains.physical_groups.curves['air_inf']

    def setMeshOptions(self):
        """
            Meshes the generated domain
        """
        gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
        gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
        gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
        if not self.data.magnet.mesh.default_mesh:
            # gmsh.option.setNumber("Mesh.AngleToleranceFacetOverlap", self.data.one_lab.mesh.AngleToleranceFacetOverlap)
            # gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", self.data.one_lab.mesh.MeshSizeFromCurvature)
            # gmsh.option.setNumber("Mesh.MeshSizeMin", self.data.one_lab.mesh.MeshSizeMin)
            # gmsh.option.setNumber("Mesh.MeshSizeMax", self.data.one_lab.mesh.MeshSizeMax)
            gmsh.option.setNumber("Mesh.Algorithm", self.data.magnet.mesh.Algorithm)
            gmsh.option.setNumber("Mesh.Optimize", self.data.magnet.mesh.Optimize)
            gmsh.option.setNumber("Mesh.ElementOrder", self.data.magnet.mesh.ElementOrder)
        else:
            gmsh.option.setNumber("Mesh.Algorithm", 6)
            gmsh.option.setNumber("Mesh.Optimize", 1)
            gmsh.option.setNumber("Mesh.ElementOrder", 2)

    def generateMesh(self):
        self.mesh.generate(2)
        self.mesh.removeDuplicateNodes()

    def checkMeshQuality(self):
        tags = self.mesh.getElements(2)[1][0]

        self.mesh_parameters['SJ'] = min(self.mesh.getElementQualities(elementTags=tags, qualityName='minSJ'))
        self.mesh_parameters['SICN'] = min(self.mesh.getElementQualities(elementTags=tags, qualityName='minSICN'))
        self.mesh_parameters['SIGE'] = min(self.mesh.getElementQualities(elementTags=tags, qualityName='minSIGE'))
        self.mesh_parameters['Gamma'] = min(self.mesh.getElementQualities(elementTags=tags, qualityName='gamma'))
        self.mesh_parameters['nodes'] = len(self.mesh.getNodes()[0])

        # gmsh.plugin.setNumber("AnalyseMeshQuality", "JacobianDeterminant", 1)
        # gmsh.plugin.setNumber("AnalyseMeshQuality", "CreateView", 100)
        # test = gmsh.plugin.run("AnalyseMeshQuality")
        # test2 = gmsh.view.getModelData(test, test)

        # gmsh.logger.getLastError()
        # gmsh.logger.get()



import pandas as pd


class ParserDAT:

    def __init__(self, dat_file_path):
        """
        Read dat file and returns its content as object attribute .pqv (postprocessed quantity value) that is a float
        :param dat_file_path: Full path to .pos file, including file name and extension.
        :return: nothing, keeps attribute pqv (postprocessed quantity value)
        """
        pqn = 'pqn'  # postprocessed quantity name
        delimiter = ' '
        columns = ['NaN', pqn]
        df = pd.read_csv(dat_file_path, delimiter=delimiter, header=None, engine='python', names=columns, skipinitialspace=True)
        self.pqv = float(df[pqn][0])



import sys
import os
import shutil
import numpy as np
from pathlib import Path
from time import sleep
import multiprocessing
import ruamel.yaml
import warnings
import gmsh


class FilesAndFolders:
    @staticmethod
    def read_data_from_yaml(full_file_path, data_class):
        with open(full_file_path, 'r') as stream:
            yaml = ruamel.yaml.YAML(typ='safe', pure=True)
            yaml_str = yaml.load(stream)
        return data_class(**yaml_str)

    @staticmethod
    def write_data_to_yaml(full_file_path, dict_of_data_class):
        def my_represent_none(self, data):
            return self.represent_scalar('tag:yaml.org,2002:null', 'null')

        yaml = ruamel.yaml.YAML()
        yaml.default_flow_style = False
        yaml.emitter.alt_null = 'Null'
        yaml.representer.add_representer(type(None), my_represent_none)
        with open(full_file_path, 'w') as yaml_file:
            yaml.dump(dict_of_data_class, yaml_file)

    @staticmethod
    def prep_folder(folder_full_path, clear: bool = False):
        if clear:
            if os.path.exists(folder_full_path):
                shutil.rmtree(folder_full_path)  # delete directory
        if not os.path.exists(folder_full_path):
            os.makedirs(folder_full_path)  # make new directory

    @staticmethod
    def get_folder_path(folder_type, folder, ref_nr, overwrite, required_folder):
        if required_folder and not (ref_nr and overwrite):
            last_nr = 0
            for study in [x.parts[-1] for x in Path(folder).iterdir() if x.is_dir()]:
                last_nr = max(int(study[study.find('_') + 1:]), last_nr)
            if overwrite and required_folder and last_nr > 0:
                run_nr = str(last_nr)
            else:
                run_nr = str(last_nr + 1)
        else:
            run_nr = str(ref_nr)

        folder_path = os.path.join(folder, folder_type + '_' + run_nr)
        FilesAndFolders.prep_folder(folder_path, overwrite and required_folder)
        return folder_path

    @staticmethod
    def print_welcome_graphics():
        print(r"  _____ _  ___        ____ ")
        print(r"|  ___(_)/ _ \ _   _/ ___| ")
        print(r"| |_  | | | | | | | \___ \ ")
        print(r"|  _| | | |_| | |_| |___) |")
        print(r"|_|   |_|\__\_\\__,_|____/ ")
        print("")


class CheckForExceptions:
    @staticmethod
    def check_inputs(run):  # RunFiQuS()
        if run.type == 'start_from_yaml':
            if run.geometry and not run.overwrite:
                warnings.warn("Warning: Geometry folder is needed only if it has to be overwritten. Ignoring it...")
            if run.solution or run.mesh:
                warnings.warn("Warning: Mesh and Solution folders are not needed. Ignoring them...")
        elif run.type == 'geometry_only':
            if run.solution or run.mesh:
                warnings.warn("Warning: Mesh and Solution folders are not needed. Ignoring them...")
        elif run.type == 'mesh_and_solve_with_post_process':
            if not run.geometry:
                raise Exception('Full path to Geometry not provided. '
                                'Insert options -> reference_files -> geometry.')
            if run.mesh and not run.overwrite:
                warnings.warn("Warning: Mesh folder is needed only if it has to be overwritten. Ignoring it...")
            if run.solution:
                warnings.warn("Warning: Solution folder is not needed. Ignoring it...")
        elif run.type == 'mesh_only':
            if not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry.')
            if run.solution:
                warnings.warn("Warning: Solution folder is not needed. Ignoring it...")
        elif run.type == 'solve_with_post_process':
            if not run.mesh or not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry and mesh.')
            if run.solution and not run.overwrite:
                warnings.warn("Warning: Solution folder is needed only if it has to be overwritten. Ignoring it...")
        elif run.type == 'solve_only':
            if not run.mesh or not run.geometry:
                raise Exception('Full path to Mesh not provided. '
                                'Insert options -> reference_files -> geometry and mesh.')
            if run.solution and not run.overwrite:
                warnings.warn("Warning: Solution folder is needed only if it has to be overwritten. Ignoring it...")
        elif run.type == 'post_process_only':
            if not run.mesh or not run.geometry or not run.solution:
                raise Exception('Full path to Solution not provided. '
                                'Insert options -> reference_files -> geometry, mesh, and solution.')

    @staticmethod
    def check_overwrite_conditions(folder_type, folder, ref_nr):
        if ref_nr:
            run_nr = str(ref_nr)
            if not os.path.exists(os.path.join(folder, folder_type + '_' + run_nr)):
                warnings.warn(
                    f'The folder {folder_type}_{run_nr} does not exist. Creating it...')
        else:
            warnings.warn(
                f'Reference number of the folder {folder_type} not provided. '
                f'Overwriting the latest {folder_type} folder...')


class GeometricFunctions:

    @staticmethod
    def centroid(X, Y):
        """
            Computes the centroid coordinates of a non-self-intersecting closed polygon
            :param X: list of X coordinate of the vertices
            :param Y: list of y coordinate of the vertices
        """
        sum_A, sum_Cx, sum_Cy = 0, 0, 0
        for i in range(len(X)):
            index = i + 1 if i != len(X) - 1 else 0
            A = X[i] * Y[index] - X[index] * Y[i]
            sum_Cx += (X[i] + X[index]) * A
            sum_Cy += (Y[i] + Y[index]) * A
            sum_A += A
        factor = 1 / (3 * sum_A)
        return [factor * sum_Cx, factor * sum_Cy]

    @staticmethod
    def arc_center_from_3_points(a, b, c):
        """
            Computes the center coordinates of an arc passing through three points
            :param a: DataRoxieParser.Coord class object of one arc point
            :param b: DataRoxieParser.Coord class object of one arc point
            :param c: DataRoxieParser.Coord class object of one arc point
        """
        ab = [a.x - b.x, a.y - b.y]
        ac = [a.x - c.x, a.y - c.y]
        sac = [a.x * a.x - c.x * c.x, a.y * a.y - c.y * c.y]
        sba = [b.x * b.x - a.x * a.x, b.y * b.y - a.y * a.y]
        yy = (sac[0] * ab[0] + sac[1] * ab[0] + sba[0] * ac[0] + sba[1] * ac[0]) / \
             (2 * ((c.y - a.y) * ab[0] - (b.y - a.y) * ac[0]))
        xx = (sac[0] * ab[1] + sac[1] * ab[1] + sba[0] * ac[1] + sba[1] * ac[1]) / \
             (2 * ((c.x - a.x) * ab[1] - (b.x - a.x) * ac[1]))
        return [-xx, -yy]


class GmshUtils:

    def __init__(self, model_name=None, verbose=True):
        self.model_name = model_name
        self.verbose = verbose

    @staticmethod
    def initialize():
        if not gmsh.is_initialized():
            gmsh.initialize(sys.argv)
            num_threads = multiprocessing.cpu_count()
            gmsh.option.setNumber('General.NumThreads', num_threads)  # enable multithreading (this seems to be only for meshing)
            gmsh.option.setNumber('Mesh.MaxNumThreads1D', num_threads)
            gmsh.option.setNumber('Mesh.MaxNumThreads2D', num_threads)
            gmsh.option.setNumber('Mesh.MaxNumThreads3D', num_threads)

            gmsh.option.setNumber('Geometry.ToleranceBoolean', 0.0000001)
            gmsh.option.setNumber('General.Terminal', 1)
            # gmsh.model.add(self.model_name)

    def check_for_event(self):  # pragma: no cover
        action = gmsh.onelab.getString("ONELAB/Action")
        if len(action) and action[0] == "check":
            gmsh.onelab.setString("ONELAB/Action", [""])
            if self.verbose:
                print("-------------------check----------------")
            gmsh.fltk.update()
            gmsh.graphics.draw()
        if len(action) and action[0] == "compute":
            gmsh.onelab.setString("ONELAB/Action", [""])
            if self.verbose:
                print("-------------------compute----------------")
            gmsh.onelab.setChanged("Gmsh", 0)
            gmsh.onelab.setChanged("GetDP", 0)
            gmsh.fltk.update()
            gmsh.graphics.draw()
        return True

    def launch_interactive_GUI(self, close_after=-1):  # pragma: no cover
        gmsh.fltk.initialize()
        while gmsh.fltk.isAvailable() and self.check_for_event():
            gmsh.fltk.wait()
            if close_after >= 0:
                sleep(close_after)
                gmsh.fltk.finalize()
        gmsh.finalize()


class RoxieParsers:
    def __init__(self, conductor, block, xyCorner):
        self.conductor = conductor
        self.block = block
        self.xyCorner = xyCorner

    @staticmethod
    def parseMap2d(map2dFile: Path, headerLines: int = 1):
        """
            Generates array-stream of values of map2dFile
            :param map2dFile: path of map2dFile containing the content to parse
            :param headerLines: which index the header line is at - will start to read after that
        """
        # Open map2dfile
        fileContent = open(map2dFile, "r").read()
        # Split content of file in rows
        fileContentByRow = fileContent.split("\n")
        # Create array-matrix to fill in with the values of the file
        output_matrix = np.array([[None for x in range(10)] for y in range(headerLines + 1, len(fileContentByRow) - 1)],
                                 dtype=float)

        # Assign values to the matrix row by row
        for index, rowContent in enumerate(fileContentByRow):
            if index > headerLines and rowContent:  # without header
                row = rowContent.split()
                output_array = np.array([])  # create temp. array
                output_array = np.append(output_array, int(row[0]))  # strands to groups
                output_array = np.append(output_array, int(row[1]))  # strands to halfturn
                output_array = np.append(output_array, float(row[2]))  # idx
                output_array = np.append(output_array, float(row[3]) / 1e3)  # x_strands in [m]
                output_array = np.append(output_array, float(row[4]) / 1e3)  # y_strands in [m]
                output_array = np.append(output_array, float(row[5]))  # Bx
                output_array = np.append(output_array, float(row[6]))  # By
                output_array = np.append(output_array, float(row[7]) / 1e6)  # Area in [m^2]
                output_array = np.append(output_array, float(row[8]))  # I_strands
                output_array = np.append(output_array, float(row[9]))  # fill factor
                output_matrix[index - headerLines - 1] = output_array  # assign into matrix
        return output_matrix

    @staticmethod
    def parseCond2d(cond2dFile: Path):
        """
            Read input file and return list of ConductorPosition objects

            # input: fileName
            # output: conductorPositionsList

        """
        # conductorStartKeyword = "CONDUCTOR POSITION IN THE CROSS-SECTION"
        blockStartKeyword = "BLOCK POSITION IN THE CROSS-SECTION"

        fileContent = open(cond2dFile, "r").read()

        # separate rows
        fileContentByRow = fileContent.split("\n")

        # Find block definition
        for i in range(len(fileContentByRow)):
            if blockStartKeyword in fileContentByRow[i]:
                startOfBlockDefinitionIndex = i

        # separate part of the data with conductor position information
        conductorPositions = fileContentByRow[5:startOfBlockDefinitionIndex - 2]

        # drop every 5th row
        conductorPositionsFourVertices = list(conductorPositions)
        del conductorPositionsFourVertices[4::5]

        # arrange data in a list of lists
        outputConductorPositions = []
        for row in conductorPositionsFourVertices:
            rowSplitStr = row.split(',')
            rowSplitFloat = [float(elem) for elem in rowSplitStr]
            outputConductorPositions.append(rowSplitFloat)

        # arrange data from list to numpy.array
        outputConductorPositionsMatrix = np.array(outputConductorPositions)

        # input: outputConductorPositions
        # output: conductorPositionsList
        conductorPositionsList = []
        for i in range(0, len(outputConductorPositions), 4):
            out = outputConductorPositions[i]
            conductor = int(out[1])
            block = int(out[2])
            xyCorner = outputConductorPositionsMatrix[i:i + 4, 4:6]
            conductorPositionsList.append(RoxieParsers(conductor, block, xyCorner))

        return conductorPositionsList



from pydantic import BaseModel
from typing import List


class Terminal(BaseModel):
    vol_st: List[int] = None  # volume number for terminal in for straightening
    surf_st: List[int] = None  # surface number for terminal in for straightening
    vol_et: List[int] = None  # volume number for terminal in for extending
    surf_et: List[int] = None  # surface number for terminal in for extending
    lc_st: List[List[List[int]]] = None  # line connections for straightening terminals
    lc_et: List[List[List[int]]] = None  # line connections for extending terminals
    z_air: float = None
    z_add: float = None
    ndpterms: List[int] = None  # number of divisions per terminal


class Winding(BaseModel):
    names: List[str] = None  # name to use in gmsh and getdp
    t_in: Terminal = Terminal()     # Terminal in
    t_out: Terminal = Terminal()    # Terminal in


class WindingsInformation(BaseModel):
    magnet_name: str = None
    windings_avg_length: float = None
    windings: Winding = Winding()
    w_names: List[str] = None
    f_names: List[str] = None
    formers: List[str] = None
    air: str = None


class SpliterBrep(BaseModel):  # Brep file model splitter data
    magnet_name: str = None
    file_name: str = None           # full file name for the brep file
    vol_firsts: List[int] = None    # list of first volumes for the partitioned model
    vol_lasts: List[int] = None      # list of last volumes for the partitioned model



from pydantic import BaseModel
from typing import (List, Literal)


class Winding_gFiQuS(BaseModel):  # Geometry related windings _inputs
    """
        Level 2: Class for FiQuS CCT
    """
    names: List[str] = None  # name to use in gmsh and getdp
    r_wms: List[float] = None  # radius of the middle of the winding
    n_turnss: List[float] = None  # number of turns
    ndpts: List[int] = None  # number of divisions of turn, i.e. number of hexagonal elements for each turn
    ndpt_ins: List[int] = None  # number of divisions of terminals in
    ndpt_outs: List[int] = None  # number of divisions of terminals in
    lps: List[float] = None  # layer pitch
    alphas: List[float] = None  # tilt angle
    wwws: List[float] = None  # winding wire widths (assuming rectangular)
    wwhs: List[float] = None  # winding wire heights (assuming rectangular)


class Winding_s(BaseModel):  # Solution time used windings _inputs (materials and BC)
    """
        Level 2: Class for FiQuS CCT
    """
    currents: List[float] = None  # current in the wire
    sigmas: List[float] = None  # electrical conductivity
    mu_rs: List[float] = None  # relative permeability


class FQPL_g(BaseModel):  # Geometry related fqpls _inputs
    """
        Level 2: Class for FiQuS CCT
    """
    names: List[str] = []  # name to use in gmsh and getdp
    fndpls: List[int] = None  # fqpl number of divisions per length
    fwws: List[float] = None  # fqpl wire widths (assuming rectangular) for theta = 0 this is x dimension
    fwhs: List[float] = None  # fqpl wire heights (assuming rectangular) for theta = 0 this is y dimension
    r_ins: List[float] = None  # radiuses for inner diameter for fqpl (radial (or x direction for theta=0) for placing the fqpl
    r_bs: List[float] = None  # radiuses for bending the fqpl by 180 degrees
    n_sbs: List[int] = None  # number of 'bending segmetns' for the 180 degrees turn
    thetas: List[float] = None  # rotation in deg from x+ axis towards y+ axis about z axis.
    z_starts: List[str] = None  # which air boundary to start at. These is string with either: z_min or z_max key from the Air region.
    z_ends: List[float] = None  # z coordinate of loop end


class FQPL_s(BaseModel):  # Solution time used fqpls _inputs (materials and BC)
    """
        Level 2: Class for FiQuS CCT
    """
    currents: List[float] = []  # current in the wire
    sigmas: List[float] = []  # electrical conductivity
    mu_rs: List[float] = []  # relative permeability


class Former_gFiQuS(BaseModel):  # Geometry related formers _inputs
    """
        Level 2: Class for FiQuS CCT
    """
    names: List[str] = None  # name to use in gmsh and getdp
    r_ins: List[float] = None  # inner radius
    r_outs: List[float] = None  # outer radius
    z_mins: List[float] = None  # extend of former  in negative z direction
    z_maxs: List[float] = None  # extend of former in positive z direction


class Former_s(BaseModel):  # Solution time used formers _inputs (materials and BC)
    """
        Level 2: Class for FiQuS CCT
    """
    sigmas: List[float] = None  # electrical conductivity
    mu_rs: List[float] = None  # relative permeability


class Air_g(BaseModel):  # Geometry related air_region _inputs
    """
        Level 2: Class for FiQuS CCT
    """
    name: str = None  # name to use in gmsh and getdp
    sh_type: str = None  # cylinder or cuboid are possible
    ar: float = None  # if box type is cuboid a is taken as a dimension, if cylinder then r is taken
    z_min: float = None  # extend of air region in negative z direction
    z_max: float = None  # extend of air region in positive z direction


class Air_s(BaseModel):  # Solution time used air _inputs (materials and BC)
    """
        Level 2: Class for FiQuS CCT
    """
    sigma: float = None  # electrical conductivity
    mu_r: float = None  # relative permeability


class GeometryCCTFiQuS(BaseModel):
    """
        Level 2: Class for FiQuS CCT for FiQuS input
    """
    windings: Winding_gFiQuS = Winding_gFiQuS()
    fqpls: FQPL_g = FQPL_g()
    formers: Former_gFiQuS = Former_gFiQuS()
    air: Air_g = Air_g()


class MeshCCT(BaseModel):
    """
        Level 2: Class for FiQuS CCT
    """
    MaxAspectWindings: float = None  # used in transfinite mesh_generators settings to define mesh_generators size along two longer lines of hex elements of windings
    ThresholdSizeMin: float = None  # sets field control of Threshold SizeMin
    ThresholdSizeMax: float = None  # sets field control of Threshold SizeMax
    ThresholdDistMin: float = None  # sets field control of Threshold DistMin
    ThresholdDistMax: float = None  # sets field control of Threshold DistMax


class SolveCCT(BaseModel):
    """
        Level 2: Class for FiQuS CCT
    """
    windings: Winding_s = Winding_s()  # windings solution time _inputs
    fqpls: FQPL_s = FQPL_s()  # fqpls solution time _inputs
    formers: Former_s = Former_s()  # former solution time _inputs
    air: Air_s = Air_s()  # air solution time _inputs
    pro_template: str = None  # file name of .pro template file
    variables: List[str] = None  # Name of variable to post-process by GetDP, like B for magnetic flux density
    volumes: List[str] = None  # Name of volume to post-process by GetDP, line Winding_1
    file_exts: List[str] = None  # Name of file extensions to post-process by GetDP, like .pos


class PostprocCCTFiQuS(BaseModel):
    """
        Class for FiQuS CCT input file
    """
    windings_wwns: List[int] = None  # wires in width direction numbers
    windings_whns: List[int] = None  # wires in height direction numbers
    additional_outputs: List[str] = None  # Name of software specific input files to prepare, like :LEDET3D
    winding_order: List[int] = None
    fqpl_export_trim_tol: List[float] = None  # this multiplier times winding extend gives 'z' coordinate above(below) which hexes are exported for LEDET, length of this list must match number of fqpls
    variables: List[str] = None  # Name of variable to post-process by python Gmsh API, like B for magnetic flux density
    volumes: List[str] = None  # Name of volume to post-process by python Gmsh API, line Winding_1
    file_exts: List[str] = None  # Name of file extensions o post-process by python Gmsh API, like .pos


class CCTDM(BaseModel):
    """
        Level 1: Class for FiQuS CCT
    """
    type: Literal['CCT_straight']
    geometry: GeometryCCTFiQuS = GeometryCCTFiQuS()
    mesh: MeshCCT = MeshCCT()
    solve: SolveCCT = SolveCCT()
    postproc: PostprocCCTFiQuS = PostprocCCTFiQuS()



from setuptools import setup
from setuptools import find_packages

with open("Readme.md", "r") as fh:
    long_description = fh.read()


setup(
    name='fiqus',
    version="2023.2.3",
    author="STEAM Team",
    author_email="steam-team@cern.ch",
    description="Source code for STEAM FiQuS tool",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://gitlab.cern.ch/steam/fiqus",
    keywords={'STEAM', 'FiQuS', 'CERN'},
    python_requires='>=3.8',
    package_data={'': ['CCT_template.pro', 'Multipole_template.pro']},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8"],

)



from pydantic import BaseModel
from typing import (Dict, List)


class Coord(BaseModel):
    """
        Class for coordinates
    """
    x: float = None
    y: float = None
    z: float = None


class Roll(BaseModel):
    """
        Class for roll2 transformation
    """
    coor: Coord = Coord()
    alph: float = None


class HyperHole(BaseModel):
    """
        Class for hyper holes
    """
    areas: List[str] = []


class HyperArea(BaseModel):
    """
        Class for hyper areas
    """
    material: str = None
    lines: List[str] = []


class HyperLine(BaseModel):
    """
        Class for hyper lines: lines, arcs, elliptic arcs, circles
    """
    type: str = None
    kp1: str = None
    kp2: str = None
    kp3: str = None
    arg1: float = None
    arg2: float = None
    elements: int = None


class CondPar(BaseModel):
    """
        Class for conductor parameters
    """
    wInsulNarrow: float = None
    wInsulWide: float = None
    dFilament: float = None
    dstrand: float = None
    fracCu: float = None
    fracSc: float = None
    RRR: float = None
    TupRRR: float = None
    Top: float = None
    Rc: float = None
    Ra: float = None
    fRhoEff: float = None
    lTp: float = None
    wBare: float = None
    hInBare: float = None
    hOutBare: float = None
    noOfStrands: int = None
    noOfStrandsPerLayer: int = None
    noOfLayers: int = None
    lTpStrand: float = None
    wCore: float = None
    hCore: float = None
    thetaTpStrand: float = None
    degradation: float = None
    C1: float = None
    C2: float = None
    fracHe: float = None
    fracFillInnerVoids: float = None
    fracFillOuterVoids: float = None


class Conductor(BaseModel):
    """
        Class for conductor type
    """
    conductorType: int = None
    cableGeom: str = None
    strand: str = None
    filament: str = None
    insul: str = None
    trans: str = None
    quenchMat: str = None
    T_0: float = None
    comment: str = None
    parameters: CondPar = CondPar()


class Cable(BaseModel):
    """
        Class for cable parameters
    """
    height: float = None
    width_i: float = None
    width_o: float = None
    ns: float = None
    transp: float = None
    degrd: float = None
    comment: str = None


class Quench(BaseModel):
    """
        Class for quench parameters
    """
    SCHeatCapa: float = None
    CuHeatCapa: float = None
    CuThermCond: float = None
    CuElecRes: float = None
    InsHeatCapa: float = None
    InsThermCond: float = None
    FillHeatCapa: float = None
    He: float = None
    comment: str = None


class Transient(BaseModel):
    """
        Class for transient parameters
    """
    Rc: float = None
    Ra: float = None
    filTwistp: float = None
    filR0: float = None
    fil_dRdB: float = None
    strandfillFac: float = None
    comment: str = None


class Strand(BaseModel):
    """
        Class for strand parameters
    """
    diam: float = None
    cu_sc: float = None
    RRR: float = None
    Tref: float = None
    Bref: float = None
    Jc_BrTr: float = None
    dJc_dB: float = None
    comment: str = None


class Filament(BaseModel):
    """
        Class for filament parameters
    """
    fildiao: float = None
    fildiai: float = None
    Jc_fit: str = None
    fit: str = None
    comment: str = None


class Insulation(BaseModel):
    """
        Class for insulation parameters
    """
    radial: float = None
    azimut: float = None
    comment: str = None


class Block(BaseModel):
    """
        Class for block list
    """
    type: int = None
    nco: int = None
    radius: float = None
    phi: float = None
    alpha: float = None
    current: float = None
    condname: str = None
    n1: int = None
    n2: int = None
    imag: int = None
    turn: float = None
    coil: int = None
    pole: int = None
    layer: int = None
    winding: int = None
    shift2: Coord = Coord()
    roll2: Roll = Roll()


class Group(BaseModel):
    """
        Class for group list
    """
    symm: int = None
    typexy: int = None
    blocks: List[int] = []  # map


class Trans(BaseModel):
    """
        Class for transformation list
    """
    x: float = None
    y: float = None
    alph: float = None
    bet: float = None
    string: str = None
    act: int = None
    bcs: List[int] = []  # map


class Iron(BaseModel):
    """
        Class for the iron yoke data
    """
    key_points: Dict[str, Coord] = {}
    hyper_lines: Dict[str, HyperLine] = {}
    hyper_areas: Dict[str, HyperArea] = {}
    hyper_holes: Dict[int, HyperHole] = {}


class Cadata(BaseModel):
    """
        Class for the conductor data
    """
    insul: Dict[str, Insulation] = {}
    filament: Dict[str, Filament] = {}
    strand: Dict[str, Strand] = {}
    transient: Dict[str, Transient] = {}
    quench: Dict[str, Quench] = {}
    cable: Dict[str, Cable] = {}
    conductor: Dict[str, Conductor] = {}


class Coil(BaseModel):
    """
        Class for the coil data
    """
    blocks: Dict[str, Block] = {}
    groups: Dict[str, Group] = {}
    transs: Dict[str, Trans] = {}


class StrandGroup(BaseModel):
    """
        Class for strand group
    """
    strand_positions: Dict[int, Coord] = {}


class Corner(BaseModel):
    """
        Class for corner positions
    """
    iL: Coord = Coord()  # inner left
    iR: Coord = Coord()  # inner right
    oL: Coord = Coord()  # outer left
    oR: Coord = Coord()  # outer right


class HalfTurnCorner(BaseModel):
    """
        Class for corner type
    """
    insulated: Corner = Corner()
    bare: Corner = Corner()


class HalfTurn(BaseModel):
    """
        Class for half-turn data
    """
    corners: HalfTurnCorner = HalfTurnCorner()
    strand_groups: Dict[int, StrandGroup] = {}


class CenterShift(BaseModel):
    """
        Class for bore center shift
    """
    inner: Coord = Coord()
    outer: Coord = Coord()


class Wedge(BaseModel):
    """
        Class for wedge positions
    """
    corners: Corner = Corner()
    corrected_center: CenterShift = CenterShift()
    coil: int = None


class BlockData(BaseModel):
    """
        Class for block data
    """
    block_corners: Corner = Corner()
    current_sign: int = None
    half_turns: Dict[int, HalfTurn] = {}


class WindingData(BaseModel):
    """
        Class for winding data
    """
    blocks: Dict[int, BlockData] = {}
    conductor_name: str = None
    conductors_number: int = None


class Winding(BaseModel):
    """
        Class for windings
    """
    windings: Dict[int, WindingData] = {}


class Layer(BaseModel):
    """
        Class for winding layers
    """
    layers: Dict[int, Winding] = {}


class Pole(BaseModel):
    """
        Class for poles
    """
    poles: Dict[int, Layer] = {}
    bore_center: Coord = Coord()


class Order(BaseModel):
    """
        Class for electrical order (block location)
    """
    coil: int = None
    pole: int = None
    layer: int = None
    winding: int = None
    block: int = None


class CoilData(BaseModel):
    """
        Class for coils
    """
    coils: Dict[int, Pole] = {}
    electrical_order: List[Order] = []


class RoxieRawData(BaseModel):
    """
        Class for the raw data
    """
    cadata: Cadata = Cadata()
    coil: Coil = Coil()


class RoxieData(BaseModel):
    """
        Class for the roxie parser
    """
    iron: Iron = Iron()
    coil: CoilData = CoilData()
    wedges: Dict[int, Wedge] = {}



import os
import pathlib

import gmsh

from fiqus.pro_assemblers.ProAssembler import ASS_PRO as aP
from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.data.RegionsModelFiQuS import RegionsModel as rM


class AssignNaming:
    def __init__(self, data: dF.FDM() = None):
        """
        Class to assign naming convention
        :param data: FiQuS data model
        """
        self.data: dF.FDM() = data

        self.naming_conv = {'omega': 'Omega', 'boundary': 'Bd_', 'powered': '_p', 'induced': '_i', 'air': '_a',
                            'air_far_field': '_aff', 'iron': '_bh', 'conducting': '_c', 'terms': 'Terms'}
        self.data.magnet.postproc.volumes = \
            [self.naming_conv['omega'] + self.naming_conv[var] if not var == 'omega'
             else self.naming_conv['omega'] for var in self.data.magnet.postproc.volumes]


class RunGetdpMultipole:
    def __init__(self, data: AssignNaming = None, sett: dF.FiQuSSettings() = None, solution_folder: str = None,
                 settings: dict = None, verbose: bool = False):
        """
        Class to solve pro file
        :param data: FiQuS data model
        :param sett: settings data model
        :param verbose: If True more information is printed in python console.
        """
        self.data: dF.FDM() = data.data
        self.naming_conv: dict = data.naming_conv
        self.set: dF.FiQuSSettings() = sett
        self.solution_folder = solution_folder
        self.settings = settings
        self.verbose: bool = verbose

        self.rm = rM()

        self.gu = GmshUtils(self.solution_folder, self.verbose)
        self.gu.initialize()
        self.occ = gmsh.model.occ
        self.mesh = gmsh.model.mesh

        self.brep_iron_curves = {1: set(), 2: set(), 3: set(), 4: set()}
        self.mesh_folder = os.path.dirname(self.solution_folder)
        self.model_file = os.path.join(self.solution_folder, 'Center_line.csv')

        self.II = (self.set.Model_Data_GS.general_parameters.I_ref[0] if self.data.magnet.postproc.compare_to_ROXIE
                   else self.data.magnet.solve.I_initial[0])

    def loadRegionFile(self):
        self.rm = Util.read_data_from_yaml(f"{os.path.join(self.mesh_folder, self.data.general.magnet_name)}.reg", rM)

    def assemblePro(self):
        self.rm.powered.vol.currents = []
        for name in self.rm.powered.vol.names:
            if name[-3:] == 'pos':
                self.rm.powered.vol.currents.append(self.II)
            else:
                self.rm.powered.vol.currents.append(-self.II)

        ap = aP(file_base_path=os.path.join(self.solution_folder, self.data.general.magnet_name),
                naming_conv=self.naming_conv)
        BH_curves_path = os.path.join(pathlib.Path(os.path.dirname(__file__)).parent, 'pro_material_functions', 'ironBHcurves.pro')
        ap.assemble_combined_pro(template=self.data.magnet.solve.pro_template, rm=self.rm, dm=self.data.magnet, BH_curves_path=BH_curves_path)

    def solve_and_postprocess(self):
        command = "-solve -v2 -pos"
        self._run(command=command)

    def postprocess(self):
        command = "-v2 -pos"
        self._run(command=command)

    def _run(self, command):
        gmsh.onelab.run(f"{self.data.general.magnet_name}",
                        f"{self.settings['GetDP_path']} "
                        f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}.pro "
                        f"{command} -msh {os.path.join(self.mesh_folder, self.data.general.magnet_name)}.msh")
        gmsh.onelab.setChanged("GetDP", 0)
        # view_tag = gmsh.view.getTags()  # this should be b
        # # # v = "View[" + str(gmsh.view.getIndex('b')) + "]"
        # gmsh.view.write(view_tag, f"{os.path.join(self.solution_folder, self.data.general.magnet_name)}-view.msh")

    def ending_step(self, gui: bool = False):
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            gmsh.clear()
            gmsh.finalize()
